#! /usr/local/bin/perl

# Copyright (C) 2006 Paul Thomas.
# This file may be copied and redistributed freely, without advance permission,
# provided that this Copyright statement is reproduced with each copy.

# LIMITATION OF WARRANTY
# NOTHING IN THIS AGREEMENT WILL BE CONSTRUED AS A REPRESENTATION MADE OR
# WARRANTY GIVEN BY PAUL THOMAS OR ANY THIRD PARTY THAT THE USE OF
# DATA PROVIDED HEREUNDER WILL NOT INFRINGE ANY PATENT, COPYRIGHT, TRADEMARK
# OR OTHER RIGHTS OF ANY THIRD PARTY. DATA IS PROVIDED "AS IS" WITHOUT
# WARRANTY OF ANY KIND WHATSOEVER, EXPRESS OR IMPLIED, INCLUDING IMPLIED
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. PAUL
# THOMAS MAKES NO WARRANTY THAT ITS DATA DOES NOT CONTAIN ERRORS.

# necessary libraries
use strict;

# default values
my $GAP_PRIOR=0.0001; ## to avoid log of (0)
my ($alignment_file,$wild_AA,$sub_AA,$pos,$priorFile,$outFile,$outWtFile,$outAlignFile,$do_clean_alignment,$is_adjWts);
my $tmpdirPath = "/tmp";

# get command-line arguments
use Getopt::Long;
$Getopt::Long::ignorecase=0; #case sensitive
&GetOptions 
(
 "h" => sub {&usage()},             # -h for help
 "a=s" =>\$alignment_file,          # -a for (a)alignment file
 "w=s" =>\$wild_AA,                 # -w for (w)ild type AA
 "m=s" =>\$sub_AA,                  # -m for (m)utant AA
 "T=s" => \$tmpdirPath,             # -T for path to (T)mp directory
 "c=i" =>\$pos,                     # -c for (c)olumn
 "A"   =>\$is_adjWts,               # -a for (A)weights
 "P:s" =>\$priorFile,               # -P for (P)rior file
 "o:s" =>\$outFile,                 # -o for (o)utput file (redirect STDOUT)
 "W:s" =>\$outWtFile,               # -w for (w)ts file to keep it  
 "C"   =>\$do_clean_alignment,      # -C for (C)leaning alignment
 "O:s" =>\$outAlignFile,            # -O for keeping cleaned (O)UT alignment file
);


&usage("tmp directory $tmpdirPath cannot be found\n") unless (-d $tmpdirPath);

$priorFile="$ENV{'PANTHER_HOME'}/data/sam_priors/uprior.9comp" unless($priorFile);
$outWtFile = "$tmpdirPath/tmp_wts.$$" unless ($outWtFile);



if ($outFile) {
  open (STDOUT, ">$outFile") || die "Cannot open $outFile\n";
}
unless (-s $priorFile) {
  die "Cannot access prior file\n";
}

#######################################################
###  MAIN BEGINS HERE
#######################################################

## GLOBALS
my ($alpha_GLOB,$sum_alpha_GLOB,$seq_GLOB,$num_components_GLOB,$mixture_GLOB);

my @alphabet=qw(A C D E F G H I K L M N P Q R S T V W Y);
my %aa2num=qw(A 0 C 1 D 2 E 3 F 4 G 5 H 6 I 7 K 8 L 9 M 10 N 11 P 12 Q 13 R 14 S 15 T 16 V 17 W 18 Y 19 - 20);
&read_prior($priorFile);
my $score=&get_score($alignment_file,$pos,$wild_AA,$sub_AA,$outAlignFile);

print "\nSCORE\t$score\n";

#######################################################
###  SUBROUTINES BEGIN HERE
#######################################################


sub read_prior(){
my $prior=shift;

open PRIOR, "<$prior" or die "Can't open prior file $prior";

my $num;
while(<PRIOR>){
  if(/Number= (\d+)/){
     $num=$1;
   }
  elsif(/Mixture= (\S+)/){
     $mixture_GLOB->[$num]=$1;
   }
  elsif(/Alpha= (.+)$/){
    my $tmp=$1;
    @{$alpha_GLOB->[$num]}=grep /\S+/, split /\s+/,$tmp;
    $sum_alpha_GLOB->[$num]=shift(@{$alpha_GLOB->[$num]});  ## AA enumeration runs from 1:20
  }
}

$num_components_GLOB=$num+1;

}

############################################

sub get_score() {
  my $alignment_file=shift;
  my $pos=shift;
  my $aa_wild=shift;
  my $aa_sub=shift;
  my $outAlignFile = shift;

  my ($a_file, $curr_id);

  open ALIGNMENT_FILE, "<$alignment_file" or die "Can't open alignment $alignment_file";
  while(<ALIGNMENT_FILE>){
    chomp;
    s/\cM//g; # remove ^M
    if(/^>Subfamily:(SF\d+)/){
      next
    }
    elsif(/^>([^\s,]+)/){
      $curr_id=$1;
    }
    elsif(/^(\S+)/){
      my $line=$1;
      $line=~s/[a-z]//g;
      $line=~s/\.//g;
      $seq_GLOB->{$curr_id}.=$line;
    }
  }
  close(ALIGNMENT_FILE);
  
  ### IF NEEDED: tmp file to strip out fasta line &&  select sequences without mut_aa

if($do_clean_alignment){ 
  if (defined($outAlignFile)) {
    $a_file = $outAlignFile;
  } else {
    $a_file="$tmpdirPath/align_a2d.$$";          
  }
  open A_FILE,">$a_file" or die "Can't open $a_file";
  foreach my $id(keys %{$seq_GLOB}){
      print A_FILE ">$id\n";
      print A_FILE "$seq_GLOB->{$id}\n";
    
  }
  close(A_FILE);
}
else{ ## fast version when alignment_file is clean;
  $a_file=$alignment_file;
}  
 my $cmd_seq_wt="seq_wt -i $a_file>$outWtFile";
 `$cmd_seq_wt` unless (-e "$outWtFile");
 
  unlink($a_file) if (!$outAlignFile && $do_clean_alignment);
  
  ## READ WEIGHTS
  open WEIGHTS_IN, "<$outWtFile" or die "CAn't open $outWtFile";
  my $line_num=0;

  my ( $weight_GLOB,$nic_GLOB,$wtd_cnt_GLOB);

  while(<WEIGHTS_IN>){
    $line_num++;
    next if ($line_num<=2);
    if(/^(\S+)\s+(\S+)/){
      $weight_GLOB->{$1}=$2 ; ## BETY ADDED */$nic
      $nic_GLOB+=$2;
    }
    
  }
  close(WEIGHTS_IN);

  unlink $outWtFile if ($outWtFile);

  ## READ_ALIGNMENT at pos $pos
  ## open (ALIGNMENT_IN, "<$a_file") or die "Can't open $a_file";
  
  for(my $aa=0;$aa<=20;$aa++){
    $wtd_cnt_GLOB->[$aa]=0;
  }


  my $nic_partial= 0;

  foreach my $id(keys %{$seq_GLOB}){
     my $aa = substr($seq_GLOB->{$id},$pos-1,1);
     my $info1 = $aa2num{$aa};
     ## this was in the OLD version without Weights adjustment $wtd_cnt_GLOB->[$info1] += $weight_GLOB->{$id};
     $nic_partial+=$weight_GLOB->{$id} if ($aa ne "-");;
  }

  ################################################3
  ############### get PROFILE
  ## my $X;
  my $tmp;
  my ($print_adjWts, $nic_actual, $posterior_mixture_GLOB,$profile_GLOB);
  
  if(!$is_adjWts){ ## if weights are not adjusted
     $print_adjWts="Weights are unadjusted: ";
     $nic_actual=$nic_partial;
     foreach my $id(keys %{$seq_GLOB}){
        my $aa = substr($seq_GLOB->{$id},$pos-1,1);
        my $info1 = $aa2num{$aa};
        $wtd_cnt_GLOB->[$info1] += $weight_GLOB->{$id};
      }
    }
  else{  ## weights are adjusted as in GHMM UPL5
    $print_adjWts="Weights are adjusted: ";
    $nic_actual=$nic_partial*(1-1./$nic_GLOB)+1;
    foreach my $id(keys %{$seq_GLOB}){
       my $aa = substr($seq_GLOB->{$id},$pos-1,1);
       next if ($aa eq "-");
       my $info1 = $aa2num{$aa};
       $wtd_cnt_GLOB->[$info1]+= $weight_GLOB->{$id}*(1.-1./$nic_GLOB +1/$nic_partial);
     }
   }

  $posterior_mixture_GLOB=&get_mixture_coefficients($wtd_cnt_GLOB,$mixture_GLOB,$alpha_GLOB,$sum_alpha_GLOB);

  for(my $comp=0;$comp<$num_components_GLOB;$comp++){
    for(my $aa=0;$aa<20;$aa++){
        $profile_GLOB->[$aa]+=$posterior_mixture_GLOB->[$comp]*( $wtd_cnt_GLOB->[$aa]+$alpha_GLOB->[$comp][$aa])/($nic_actual+$sum_alpha_GLOB->[$comp]);
      }
    }


  for(my $comp=0;$comp<$num_components_GLOB;$comp++){
    print "$print_adjWts\t$comp\t$posterior_mixture_GLOB->[$comp]\n";
  }
  print "\n";

  for(my $aa=0;$aa<20;$aa++){  
    print "$print_adjWts\t$alphabet[$aa]\t$profile_GLOB->[$aa]\n";
  }
  print "\n";
  #get probability for wild type and mutant
  my $p_wild = $profile_GLOB->[$aa2num{$aa_wild}] ;
  my $p_sub = $profile_GLOB->[$aa2num{$aa_sub}] ;;
  die "******** p_sub is 0" if ($p_sub== 0.);
  
  my $score=log($p_sub/$p_wild);
  print "P_wild=$p_wild\nP_sub=$p_sub\nNIC=$nic_GLOB\nNIC_actual=$nic_actual\n";

  return $score;
}
###############################################################

sub get_mixture_coefficients(){

my $wtd_cnt=shift;
my $mixture=shift;
my $alpha=shift;
my $sum_alpha=shift;

my $p;
my $s;

my $tmp;

for(my $i=0;$i<$num_components_GLOB;$i++){
    for(my $aa=0;$aa<20;$aa++){
      $tmp->[$aa]=$wtd_cnt->[$aa]+$alpha->[$i][$aa];
    }
    $p->[$i]=$mixture->[$i]*exp(&lbeta($tmp)-&lbeta($alpha->[$i]));
    $s+=$p->[$i];
  }

### normalize
for(my $i=0;$i<$num_components_GLOB;$i++){
    $p->[$i]/=$s;
  }

return $p;

}


###########################################################

sub lbeta(){

   my $aa=shift;  ## array

   my $log_beta;
   my $sum;
   foreach my $a(@{$aa}){
        $log_beta+=&lgamma($a);
        $sum+=$a;
      }
  $log_beta-=&lgamma($sum);

  return $log_beta;
 }

#####################################################

sub lgamma(){

my $x=shift;

my @coef=qw( 76.18009172947146 -86.50532032941677  24.01409824083091  -1.231739572450155 .1208650973866179e-2 -.5395239384953e-5);

my $stp = 2.5066282746310005;

$x--;
my $tmp=$x+5.5;
$tmp=($x+.5)*log($tmp)-$tmp;
my $ser=1.000000000190015;

for (my $j=0;$j<6;$j++){
  $x+=1.;
  $ser+=$coef[$j]/$x;
}

my $res=$tmp+log($stp*$ser);

return $res;


}
